from fixture import make_test


def test_waveform_basic(make_test):
    make_test("waveform_0")
    make_test("waveform_1")
    make_test("waveform_1n")
    make_test("waveform_2")
    make_test("waveform_3")
    make_test("waveform_4")
    make_test("waveform_5")
    make_test("waveform_6")
    make_test("waveform_7")
    make_test("waveform_8")
    make_test("waveform_9")
    make_test("waveform_10")
    make_test("waveform_11")
    make_test("waveform_12")


def test_waveform_sub_cycle(make_test):
    make_test("waveform_sub_cycle")
    make_test("waveform_sub_cycle_gaps")



# The MIT License (MIT)
#
# Copyright (c) 2011-2018 Aliaksei Chapyzhenka, BreizhGeek, Kazuki Yamamoto,
#                         Stefan Wallentowitz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Translated to Python from original file:
# https://github.com/drom/wavedrom/blob/master/src/WaveDrom.js
#

from collections import namedtuple
import svgwrite

from .base import SVGBase


class RenderState:
    def __init__(self, x=0, y=0, xmax=0):
        self.x = x
        self.y = y
        self.xmax = xmax

    def __str__(self):
        return "x={} y={}, xmax={}".format(self.x, self.y, self.xmax)


RenderObject = namedtuple("RenderObject", "name x y")


class Assign(SVGBase):
    def render_tree(self, tree, state):
        state.xmax = max(state.xmax, state.x)
        y = state.y
        for i in range(1, len(tree)):
            if isinstance(tree[i], list):
                state = self.render_tree(tree[i], RenderState(x=state.x+1, y=state.y, xmax=state.xmax))
            else:
                tree[i] = RenderObject(name=tree[i], x=state.x+1, y=state.y)
                state.y += 2
        tree[0] = RenderObject(name=tree[0], x=state.x, y=round((y+state.y-2)/2))
        state.x -= 1

        return state

    def draw_body(self, type, ymin, ymax):
        circle = ' M 4,0 C 4,1.1 3.1,2 2,2 0.9,2 0,1.1 0,0 c 0,-1.1 0.9,-2 2,-2 1.1,0 2,0.9 2,2 z'
        gates = {
            '~':  'M -11,-6 -11,6 0,0 z m -5,6 5,0' + circle,
            '=':  'M -11,-6 -11,6 0,0 z m -5,6 5,0',
            '&':  'm -16,-10 5,0 c 6,0 11,4 11,10 0,6 -5,10 -11,10 l -5,0 z',
            '~&': 'm -16,-10 5,0 c 6,0 11,4 11,10 0,6 -5,10 -11,10 l -5,0 z' + circle,
            '|':  'm -18,-10 4,0 c 6,0 12,5 14,10 -2,5 -8,10 -14,10 l -4,0 c 2.5,-5 2.5,-15 0,-20 z',
            '~|': 'm -18,-10 4,0 c 6,0 12,5 14,10 -2,5 -8,10 -14,10 l -4,0 c 2.5,-5 2.5,-15 0,-20 z' + circle,
            '^':  'm -21,-10 c 1,3 2,6 2,10 m 0,0 c 0,4 -1,7 -2,10 m 3,-20 4,0 c 6,0 12,5 14,10 -2,5 -8,10 -14,10 l -4,0 c 1,-3 2,-6 2,-10 0,-4 -1,-7 -2,-10 z',
            '~^': 'm -21,-10 c 1,3 2,6 2,10 m 0,0 c 0,4 -1,7 -2,10 m 3,-20 4,0 c 6,0 12,5 14,10 -2,5 -8,10 -14,10 l -4,0 c 1,-3 2,-6 2,-10 0,-4 -1,-7 -2,-10 z' + circle,
            '+':  'm -8,5 0,-10 m -5,5 10,0 m 3,0 c 0,4.418278 -3.581722,8 -8,8 -4.418278,0 -8,-3.581722 -8,-8 0,-4.418278 3.581722,-8 8,-8 4.418278,0 8,3.581722 8,8 z',
            '*':  'm -4,4 -8,-8 m 0,8 8,-8 m 4,4 c 0,4.418278 -3.581722,8 -8,8 -4.418278,0 -8,-3.581722 -8,-8 0,-4.418278 3.581722,-8 8,-8 4.418278,0 8,3.581722 8,8 z'
        }
        iec = {
            "BUF": 1, "INV": 1, "AND": '&',  "NAND": '&',
            "OR": '\u22651', "NOR": '\u22651', "XOR": '=1', "XNOR": '=1', "box": ''
        }
        circled = { "INV", "NAND", "NOR", "XNOR" }

        if ymax == ymin:
            ymax = 4
            ymin = -4

        if type in gates:
            return self.element.path(class_='gate', d=gates[type])
        elif type in iec:
            g = self.container.g()
            if type in circled:
                path = self.element.path(class_="gate", d="m -16,{} 16,0 0,{} -16,0 z {}".format(ymin-3, ymax-ymin+6, circle))
            else:
                path = self.element.path(class_="gate", d="m -16,{} 16,0 0,{} -16,0 z".format(ymin-3, ymax-ymin+6))
            g.add(path)
            tspan = self.element.tspan(iec[type], x=[-14], y=[4], class_='wirename')
            text = self.element.text('')
            text.add(tspan)
            g.add(text)
            return g
        else:
            tspan = self.element.tspan(type, x=[-14], y=[4], class_='wirename')
            text = self.element.text('')
            text.add(tspan)
            return text

    def draw_gate(self, spec):
        ret = self.container.g()

        ys = [s[1] for s in spec[2:]]

        ymin = min(ys)
        ymax = max(ys)

        g = self.container.g(transform="translate(16,0)")
        g.add(self.element.path(d="M {},{} {},{}".format(spec[2][0], ymin, spec[2][0], ymax), class_='wire'))
        ret.add(g)

        for s in spec[2:]:
            path = self.element.path(d="m {},{} 16,0".format(s[0], s[1]), class_='wire')
            ret.add(self.container.g().add(path))

        g = self.container.g(transform="translate({},{})".format(spec[1][0], spec[1][1]))
        g.add(self.element.title(spec[0]))
        g.add(self.draw_body(spec[0], ymin - spec[1][1], ymax - spec[1][1]))
        ret.add(g)

        return ret

    def draw_boxes(self, tree, xmax):
        ret = self.container.g()
        spec = []

        if isinstance(tree, list):
            spec.append(tree[0].name);
            spec.append([32 * (xmax - tree[0].x), 8 * tree[0].y]);
            for t in tree[1:]:
                if isinstance(t, list):
                    spec.append([32 * (xmax - t[0].x), 8 * t[0].y])
                else:
                    spec.append([32 * (xmax - t.x), 8 * t.y])

            ret.add(self.draw_gate(spec))

            for t in tree[1:]:
                ret.add(self.draw_boxes(t, xmax))
        else:
            fname = tree.name
            fx = 32 * (xmax - tree.x)
            fy = 8 * tree.y
            g = self.container.g(transform="translate({},{})".format(fx, fy))
            g.add(self.element.title(fname))
            g.add(self.element.path(d='M 2,0 a 2,2 0 1 1 -4,0 2,2 0 1 1 4,0 z'))
            tspan = self.element.tspan(fname, x=[-4], y=[4], class_='pinname')
            text = self.element.text('')
            text.add(tspan)
            g.add(text)
            ret.add(g)

        return ret

    def render(self, index = 0, source = {}, output = []):
        STYLE = ".pinname {font-size:12px; font-style:normal; font-variant:normal; font-weight:500; font-stretch:normal; text-align:center; text-anchor:end; font-family:Helvetica} .wirename {font-size:12px; font-style:normal; font-variant:normal; font-weight:500; font-stretch:normal; text-align:center; text-anchor:start; font-family:Helvetica} .wirename:hover {fill:blue} .gate {color:#000; fill:#ffc; fill-opacity: 1;stroke:#000; stroke-width:1; stroke-opacity:1} .gate:hover {fill:red !important; } .wire {fill:none; stroke:#000; stroke-width:1; stroke-opacity:1} .grid {fill:#fff; fill-opacity:1; stroke:none}"

        tree = source.get("assign")
        state = RenderState(x=0, y=2, xmax=0)

        for t in tree:
            state = self.render_tree(t, state)
            state.x += 1

        xmax = state.xmax + 3

        width = 32 * (xmax + 1) + 1
        height = 8 * (state.y + 1) - 7
        ilen = 4 * (xmax + 1)
        jlen = state.y + 1

        grid = self.container.g()

        for i in range(ilen+1):
            for j in range(jlen+1):
                grid.add(self.element.rect(height=1, width=1, x=(i * 8 - 0.5), y=(j * 8 - 0.5), class_='grid'))

        for t in tree:
            content = self.draw_boxes(t, xmax)


        attr = { 'viewBox': "0 0 {} {}".format(width, height)}
        template = svgwrite.Drawing(id="svgcontent_{index}".format(index=index), size=[width, height], **attr)
        template.defs.add(svgwrite.container.Style(content=STYLE))
        g = self.container.g(transform="translate(0.5,0.5)")
        g.add(grid)
        g.add(content)
        template.add(g)
        return template



import pytest

from wavedrom import WaveDrom
from regressions import all

@pytest.fixture
def make_regression():
    def _make_regression(test):
        w = WaveDrom()
        w.lane.period = test.period
        w.lane.hscale = test.hscale
        w.lane.phase = test.phase
        output = w.parse_wave_lane(test.wave, int(test.period * test.hscale - 1))
        assert(output == test.expected)


    return _make_regression

def test_regression(make_regression):
    for test in all:
        make_regression(test)


from fixture import make_test


def test_assign(make_test):
    make_test("assign_xor")
    make_test("assign_gray2binary")
    make_test("assign_binary2gray")
    make_test("assign_74ls688")
    make_test("assign_iec60617")



# The MIT License (MIT)
#
# Copyright (c) 2011-2018 Aliaksei Chapyzhenka, BreizhGeek, Kazuki Yamamoto,
#                         Stefan Wallentowitz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Translated to Python from original file:
# https://github.com/drom/wavedrom/blob/master/src/WaveDrom.js
#

import sys
if sys.version_info < (3, 0):
    from HTMLParser import HTMLParser
else:
    from html.parser import HTMLParser
from math import floor

import svgwrite

from .base import SVGBase


class Options:
    def __init__(self, vspace=80, hspace=640, lanes=2, bits=32, fontsize=14, bigendian=False, fontfamily='sans-serif',
                 fontweight='normal'):
        self.vspace = vspace if vspace > 19 else 80
        self.hspace = hspace if hspace > 39 else 640
        self.lanes = lanes if lanes > 0 else 2
        self.bits = bits if bits > 4 else 32
        self.fontsize = fontsize if fontsize > 5 else 14
        self.bigendian = bigendian
        self.fontfamily = fontfamily
        self.fontweight = fontweight


def type_style(t):
    if t==2:
        return ';fill:hsl(0,100%,50%)'
    elif t==3:
        return ';fill:hsl(80,100%,50%)'
    elif t==4:
        return ';fill:hsl(170,100%,50%)'
    elif t==5:
        return ';fill:hsl(45,100%,50%)'
    elif t==6:
        return ';fill:hsl(126,100%,50%)'
    elif t==7:
        return ';fill:hsl(215,100%,50%)'
    else:
        return ''


class TspanParser(HTMLParser, object):
    tags = {
        'o': {'text_decoration': 'overline'},
        'ins': {'text_decoration': 'underline'},
        'sub': {'baseline_shift': 'sub'},
        'sup': {'baseline_shift': 'super'},
        'b': {'font_weight': 'bold'},
        'i': {'font_style': 'italic'},
        's': {'text_decoration': 'line-through'},
        'tt': {'font_family': 'monospace'},
    }

    def __init__(self):
        super(TspanParser, self).__init__()
        self.text = []
        self.state = []

    def handle_starttag(self, tag, attrs):
        self.state.append(tag)

    def handle_endtag(self, tag):
        if self.state.pop() != tag:
            raise RuntimeError("Unexpected closing tag: {}".format(tag))

    def get_style(self):
        return {k: v for d in [self.tags[t] for t in self.state] for k, v in d.items()}

    def handle_data(self, data):
        if len(self.state) == 0:
            self.text.append(svgwrite.text.TSpan(data))
        else:
            self.text.append(svgwrite.text.TSpan(data, **self.get_style()))

    def get_text(self):
        return self.text


class BitField(SVGBase):
    def tspan_parse(self, text):
        parser = TspanParser()
        parser.feed(text)
        return parser.get_text()

    def hline(self, len, x=0, y=0):
        return self.element.line(start=(x,y), end=(x+len,y))

    def vline(self, len, x=0, y=0):
        return self.element.line(start=(x,y), end=(x,y+len))

    def get_text(self, body, x, y=None):
        font = {"font_size": self.opt.fontsize,
                "font_family": self.opt.fontfamily,
                "font_weight": self.opt.fontweight,}
        x_list = None
        if x:
            x_list = [x]
        y_list = None
        if y:
            y_list = [y]
        text = self.element.text('', x=x_list, y=y_list, **font)
        for t in self.tspan_parse(str(body)):
            text.add(t)
        return text

    def get_label(self, attr, x, y, step=0, length=0):
        return self.get_text(attr, x, y)

    def get_attr(self, e, step, lsbm, msbm):
        x = [step * (self.mod - ((msbm + lsbm) / 2) - 1)]
        attr = e['attr']
        bits = e['bits']
        return self.get_label(attr, x, 0, step, bits)

    def labelArr(self, desc):
        step = self.opt.hspace / self.mod
        bits = self.container.g(transform="translate({},{})".format(step/2, self.opt.vspace/5))
        names = self.container.g(transform="translate({},{})".format(step/2, self.opt.vspace/2+4))
        attrs = self.container.g(transform="translate({},{})".format(step/2, self.opt.vspace))
        blanks = self.container.g(transform="translate(0,{})".format(self.opt.vspace/4))

        for e in desc:
            lsbm = 0
            msbm = self.mod - 1
            lsb = self.index * self.mod
            msb = (self.index + 1) * self.mod - 1

            if floor(e["lsb"] / self.mod) == self.index:
                lsbm = e["lsbm"]
                lsb = e["lsb"]
                if floor(e["msb"] / self.mod) == self.index:
                    msb = e["msb"]
                    msbm = e["msbm"]
            else:
                if floor(e["msb"] / self.mod) == self.index:
                    msb = e["msb"]
                    msbm = e["msbm"]
                else:
                    continue

            bits.add(self.get_text(lsb, x=[step*(self.mod-lsbm - 1)]))
            if lsbm != msbm:
                bits.add(self.get_text(msb, x=[step * (self.mod - msbm - 1)]))
            if e.get('name'):
                x = step*(self.mod-((msbm+lsbm)/2)-1)
                names.add(self.get_label(e['name'], x, 0))

            if not e.get('name') or e.get('type'):
                style = 'fill-opacity:0.1' + type_style(e.get('type', 0))
                insert = [step * (self.mod - msbm - 1), 0]
                size = [step * (msbm - lsbm + 1), self.opt.vspace/2]
                blanks.add(self.element.rect(insert=insert, size=size, style=style))
            if e.get('attr'):
                attrs.add(self.get_attr(e, step, lsbm, msbm))

        g = self.container.g()
        g.add(blanks)
        g.add(bits)
        g.add(names)
        g.add(attrs)
        return g

    def labels(self, desc):
        g = self.container.g(text_anchor='middle')
        g.add(self.labelArr(desc))
        return g

    def cage(self, desc):
        hspace = self.opt.hspace
        vspace = self.opt.vspace
        mod = self.mod

        g = self.container.g(stroke='black', stroke_width=1, stroke_linecap='round', transform="translate(0,{})".format(vspace/4))

        g.add(self.hline(hspace));
        g.add(self.vline(vspace / 2));
        g.add(self.hline(hspace, 0, vspace / 2));

        i = self.index * mod
        for j in range(mod, 0, -1):
            if j == mod or any([(e["lsb"] == i) for e in desc]):
                g.add(self.vline((vspace / 2), j * (hspace / mod)));
            else:
                g.add(self.vline((vspace / 16), j * (hspace / mod)));
                g.add(self.vline((vspace / 16), j * (hspace / mod), vspace * 7 / 16));
            i += 1

        return g

    def lane(self, desc):
        g = self.container.g(transform = "translate({},{})".format(4.5, (self.opt.lanes-self.index-1)*self.opt.vspace + 0.5))
        g.add(self.cage(desc))
        g.add(self.labels(desc))
        return g

    def render(self, desc, opt = Options()):
        self.opt = opt

        width = opt.hspace + 9
        height = opt.vspace * opt.lanes + 5
        viewbox = { 'viewBox': "0 0 {} {}".format(width, height)}

        template = svgwrite.Drawing(id="svgcontent", size=[width, height], **viewbox)

        lsb = 0
        self.mod = int(opt.bits / opt.lanes)

        for e in desc:
            e["lsb"] = lsb
            e["lsbm"] = lsb % self.mod
            lsb += e['bits']
            e['msb'] = lsb - 1
            e['msbm'] = e['msb'] % self.mod

        for i in range(opt.lanes):
            self.index = i
            template.add(self.lane(desc))

        return template

    def renderJson(self, source):
        opt = Options()
        if source.get("options"):
            opt = Options(**source['options'])
        if source.get("reg"):
            return self.render(source['reg'], opt)
